// -*- C++ -*-
/*!
 * @file  EchoReader.cpp
 * @brief EchoReader component
 * @date 2009/07/17
 * @author Kazuo Nakayoshi <kazuo.nakayoshi@kek.jp>
 *
 * Copyright (C) 2009
 *     Kazuo Nakayoshi
 *     Electronics System Group,
 *     KEK, Japan.
 *     All rights reserved.
 *
 */

#include "EchoReader.h"

/// For stand alone test, comment out next line.
#define USE_OUTPORT

// Module specification
static const char* echoReader_spec[] =
{
    "implementation_id", "EchoReader",
    "type_name",         "EchoReader",
    "description",       "EchoReader component",
    "version",           "1.0",
    "vendor",            "Kazuo Nakayoshi, KEK",
    "category",          "example",
    "activity_type",     "DataFlowComponent",
    "max_instance",      "10",
    "language",          "C++",
    "lang_type",         "compile",
    ""
};

const std::string EchoReader::HOSTADDR = "127.0.0.1"; ///localhost
const unsigned int EVENT_BYTE_SIZE = 4;

EchoReader::EchoReader(RTC::Manager* manager)
    : DAQMW::MlfComponent(manager),
      //m_InPort("echoReader_in",   m_in_data),
      m_OutPort("echoReader_out", m_out_data),

      ///m_in_status(BUF_SUCCESS),
      m_out_status(BUF_SUCCESS),
      m_sock(0), 
      m_dataByteSize(400),///100 events
      m_wdata(0),
      m_rdata(0),
      m_debug(false)
{
    // Registration: InPort/OutPort/Service

    // Set InPort buffers
    ///registerInPort ("echoReader_in",   m_InPort);
    registerOutPort("echoReader_out", m_OutPort);

    init_command_port();
    init_state_table( );
    set_comp_name(COMP_NONAME);

    m_eventByteSize = EVENT_BYTE_SIZE;
}

EchoReader::~EchoReader()
{
    if(m_wdata) {
	delete [] m_wdata;
	m_wdata = 0;
    }
    if(m_rdata) {
	delete [] m_rdata;
	m_rdata = 0;
    }
}


RTC::ReturnCode_t EchoReader::onInitialize()
{
    if (m_debug) {
	std::cerr << "EchoReader::onInitialize()" << std::endl;
    }

    return RTC::RTC_OK;
}

RTC::ReturnCode_t EchoReader::onExecute(RTC::UniqueId ec_id)
{
    daq_do();
    return RTC::RTC_OK;
}

int EchoReader::daq_dummy() 
{
    return 0;
}

int EchoReader::daq_configure() 
{
    std::cerr << "*** EchoReader::configure" << std::endl;

    ::NVList* paramList;
    paramList = m_daq_service0.getCompParams();
    parse_params(paramList);

    m_rdata = new unsigned char[m_dataByteSize];
    m_wdata = new unsigned char[m_dataByteSize];
    return 0;
}

int EchoReader::parse_params(::NVList* list)
{

    std::cerr << "param list length:" << (*list).length() << std::endl;

    int length = (*list).length();
    for(int i=0; i< length ; i++) {
        if (m_debug) {
            std::cerr << "NVList[" << (*list)[i].name 
                      << ","<< (*list)[i].value << "]" << std::endl;
	   }
    }
    return 0;
}

int EchoReader::daq_unconfigure() 
{
    std::cerr << "*** EchoReader::unconfigure" << std::endl;
    return 0;
}

int EchoReader::daq_start() 
{
    std::cerr << "*** EchoReader::start" << std::endl;
 
    ///m_in_status  = BUF_SUCCESS;
    m_out_status = BUF_SUCCESS;

    try {
	/// Create socket and connect to echo server.
	m_sock = new DAQMW::Sock(); 
	m_sock->connect(HOSTADDR, PORTNO);
    } catch (DAQMW::SockException& e) {
        std::cerr << "Sock Fatal Error : " << e.what() << std::endl;
    } catch (...) {
        std::cerr << "Sock Fatal Error : Unknown" << std::endl;
    }

    return 0;
}

int EchoReader::daq_stop() 
{
    std::cerr << "*** EchoReader::stop" << std::endl;
    /// disconnect from echo server
    m_sock->disconnect(); 
    delete m_sock;        
    m_sock = 0;           

    return 0;
}

int EchoReader::daq_pause()
{
    std::cerr << "*** EchoReader::pause" << std::endl;
    return 0;
}

int EchoReader::daq_resume()
{
    std::cerr << "*** EchoReader::resume" << std::endl;
    return 0;
}

/// generates a Gaussian distribution(mean: 60, sigma: 10) number
/// using uniform random numbers.
unsigned int EchoReader::getGauss()
{
    double gdat = 0.0;
    double amp  = 10.0;
    int loop = 12; 
    for (int i = 0; i< loop; i++) {
	gdat = gdat + (double)random()/((double)RAND_MAX + 1);
    }
    return (unsigned)(gdat * amp);
}

/// send data to echo server 
int EchoReader::send_data_to_echoServer()
{
    /// generates 100 events
    for (int i = 0; i < m_dataByteSize; i+=4) {
	unsigned int gdat = getGauss();
	*(unsigned*)&m_wdata[i] = gdat;
	//*(unsigned*)&m_wdata[i] = htonl(gdat);
    }

    /// send to echo server
    int status = m_sock->writeAll(m_wdata, m_dataByteSize);
    if(status == DAQMW::Sock::ERROR_FATAL) {
        std::cerr << "fatal error..." << std::endl;
	fatal_error_report(USER_ERROR1, -1);
    }
    else if(status == DAQMW::Sock::ERROR_TIMEOUT) {
	std::cerr << "Timeout.. retry..." << std::endl;
	fatal_error_report(USER_ERROR2, -1);
    }
    return 0;
}

/// receive data from echo server
int EchoReader::recv_data_from_echoServer()
{
    /// read 100 events data from echo server
    int status = m_sock->readAll(m_rdata, m_dataByteSize);
    if(status == DAQMW::Sock::ERROR_FATAL) {
        std::cerr << "fatal error..." << std::endl;
	fatal_error_report(USER_ERROR1, -1);
    }
    else if(status == DAQMW::Sock::ERROR_TIMEOUT) {
	std::cerr << "Timeout.. retry..." << std::endl;
	fatal_error_report(USER_ERROR2, -1);
    }

    if (m_debug) {
	unsigned char* rdata_ptr = &m_rdata[0];
	for (int i = 0; i < m_dataByteSize/4; i++) {
	    unsigned int rdat = *(unsigned int*)rdata_ptr;
	    //std::cerr << ntohl(rdat) << std::endl;
	    std::cerr << rdat << std::endl;
	    rdata_ptr += 4;
	}
    }
    return 0;
}

/// set data to OutPort buffer
int EchoReader::set_data_to_OutPortBuf( unsigned int seq_num )
{
    unsigned int addr  = 0;
    unsigned int daqId = 0; 
    unsigned int modNo = 0;

    unsigned char header[8];
    unsigned char footer[8];
    
    set_header(header, daqId, modNo, m_dataByteSize);
    set_footer(footer, addr, seq_num);

    ///set OutPort buffer length
    m_out_data.data.length(m_dataByteSize + HEADER_BYTE_SIZE + FOOTER_BYTE_SIZE); 
    memcpy( &(m_out_data.data[0]), header, HEADER_BYTE_SIZE);
    memcpy( &(m_out_data.data[HEADER_BYTE_SIZE]), &m_rdata[0], m_dataByteSize);
    memcpy( &(m_out_data.data[HEADER_BYTE_SIZE + m_dataByteSize]), footer, FOOTER_BYTE_SIZE);

    return 0;
}

/// check OutPort status
int EchoReader::check_outPort_status(int status)
{
    int ret = 0;

    if (status == BUF_SUCCESS) {
        m_loop++;
	m_total_event += (m_dataByteSize/m_eventByteSize);
        if (m_debug) {
            if (m_loop%100 == 0) {
                std::cerr << "EchoReader: m_loop = " << m_loop << std::endl;
                std::cerr << "\033[A\r";
            }
        }
    }
    else if (status == BUF_TIMEOUT) {
        std::cerr << "EchoReader::Time Out occurred..." << std::endl;
    }
    else if (status == BUF_FATAL) {
        std::cerr << "### EchoReader::Fatal error occurred..." << std::endl;
        ret = -1;
    }
    return ret;
}

int EchoReader::daq_run() 
{
    if (m_debug) {
	std::cerr << "*** EchoReader::run" << std::endl;
    }

    send_data_to_echoServer();
    recv_data_from_echoServer();

#ifdef USE_OUTPORT	
    set_data_to_OutPortBuf(m_loop);
    m_out_status = m_OutPort.write(m_out_data); /// send data from OutPort

    if (check_outPort_status(m_out_status) == -1) {
	std::cerr << "### EchoReader: OutPort.write(): FATAL ERROR\n";
	fatal_error_report(USER_ERROR3, -1);
        return 0;
    }

#else
    m_loop++;
    m_total_event += (m_dataByteSize/m_eventByteSize);
#endif
    if (check_trans_lock() ) {  /// got stop command 
        set_trans_unlock();
        return 0;
    }

    usleep(100000); /// sleep for 100ms to adjust histogram update time

    return 0;
}

extern "C"
{
     void EchoReaderInit(RTC::Manager* manager)
    {
	RTC::Properties profile(echoReader_spec);
	manager->registerFactory(profile,
				 RTC::Create<EchoReader>,
				 RTC::Delete<EchoReader>);
    }
};
